import { AnnotatorArea } from '../AnnotationArea';
import { AnnotatorAreaState } from '../AnnotationState';
import { AnnotatorBase } from './AnnotatorBase';

export class AnnotatorAreaEvent {
  public annotatorArea: AnnotatorArea;
  public cancelable = false;

  private _defaultPrevented = false;
  public get defaultPrevented(): boolean {
    return this._defaultPrevented;
  }

  public preventDefault(): void {
    this._defaultPrevented = true;
  }

  constructor(annotatorArea: AnnotatorArea, cancelable = false) {
    this.annotatorArea = annotatorArea;
    this.cancelable = cancelable;
  }
}

export class AnnotatorAreaRenderEvent extends AnnotatorAreaEvent {
  public dataUrl: string;
  public state: AnnotatorAreaState;

  constructor(annotatorArea: AnnotatorArea, dataUrl: string, state: AnnotatorAreaState) {
    super(annotatorArea, false);
    this.dataUrl = dataUrl;
    this.state = state;
  }
}


export class AnnotatorEvent extends AnnotatorAreaEvent {
  public marker?: AnnotatorBase;

  constructor(annotatorArea: AnnotatorArea, marker?: AnnotatorBase, cancelable = false) {
    super(annotatorArea, cancelable);
    this.marker = marker;
  }
}

/**
 * General AnnotatorArea event handler type.
 */
export type AnnotatorAreaEventHandler = (event: AnnotatorAreaEvent) => void;

/**
 * AnnotatorArea render event handler type.
 */
export type AnnotatorAreaRenderEventHandler = (event: AnnotatorAreaRenderEvent) => void;

/**
 * Annotator event handler type.
 */
export type AnnotatorEventHandler = (event: AnnotatorEvent) => void;

/**
 * Describes a repository of AnnotatorArea event handlers.
 */
export interface IEventListenerRepository {
  /**
   * Event handlers for the `render` event.
   */
  render: AnnotatorAreaRenderEventHandler[];
  /**
   * Event handlers for the `beforeclose` event.
   */
  beforeclose: AnnotatorAreaEventHandler[];
  /**
   * Event handlers for the `close` event.
   */
  close: AnnotatorAreaEventHandler[];
  /**
   * Event handlers for the `show` event.
   */
  show: AnnotatorAreaEventHandler[];
  /**
   * Event handlers for the `restorestate` event.
   */
  restorestate: AnnotatorAreaEventHandler[];
  /**
   * Event handlers for the `statechange` event.
   * 
   * @since 2.23.0
   */
  statechange: AnnotatorAreaEventHandler[];
  /**
   * Event handlers for the `markerselect` event.
   */
  markerselect: AnnotatorEventHandler[];
  /**
   * Event handlers for the `markerdeselect` event.
   */
  markerdeselect: AnnotatorEventHandler[];
  /**
   * Event handlers for the `markercreating` event.
   */
  markercreating: AnnotatorEventHandler[];
  /**
   * Event handlers for the `markercreated` event.
   */
  markercreate: AnnotatorEventHandler[];
  /**
   * Event handlers for the `markerbeforedelete` event.
   */
  markerbeforedelete: AnnotatorEventHandler[];
  /**
   * Event handlers for the `markerdelete` event.
   */
  markerdelete: AnnotatorEventHandler[];
  /**
   * Event handlers for the `markerchange` event.
   * 
   * @since 2.23.0
   */
  markerchange: AnnotatorEventHandler[];
  /**
   * Event handlers for the `focus` event.
   * 
   * @since 2.19.0
   */
  focus: AnnotatorAreaEventHandler[];
  /**
   * Event handlers for the `blur` event.
   * 
   * @since 2.19.0
   */
  blur: AnnotatorAreaEventHandler[];
}

/**
 * Event handler type for a specific event type.
 */
export type EventHandler<
  T extends keyof IEventListenerRepository
> = T extends 'markerselect'
  ? AnnotatorEventHandler
  : T extends 'markerdeselect'
  ? AnnotatorEventHandler
  : T extends 'markercreating'
  ? AnnotatorEventHandler
  : T extends 'markercreate'
  ? AnnotatorEventHandler
  : T extends 'markerbeforedelete'
  ? AnnotatorEventHandler
  : T extends 'markerdelete'
  ? AnnotatorEventHandler
  : T extends 'markerchange'
  ? AnnotatorEventHandler
  : T extends 'render'
  ? AnnotatorAreaRenderEventHandler
  : AnnotatorAreaEventHandler;

/**
 * Event handler repository.
 */
export class EventListenerRepository implements IEventListenerRepository {
  /**
   * Event handlers for the `render` event.
   */
  render: AnnotatorAreaRenderEventHandler[] = [];
  /**
   * Event handlers for the `beforeclose` event.
   */
  beforeclose: AnnotatorAreaEventHandler[] = [];
  /**
   * Event handlers for the `close` event.
   */
  close: AnnotatorAreaEventHandler[] = [];
  /**
   * Event handlers for the `show` event.
   */
  show: AnnotatorAreaEventHandler[] = [];
  /**
   * Event handlers for the `restorestate` event.
   */
  restorestate: AnnotatorAreaEventHandler[] = [];
  /**
   * Event handlers for the `statechange` event.
   * 
   * @since 2.23.0
   */
  statechange: AnnotatorAreaEventHandler[] = [];
  /**
   * Event handlers for the `markerselect` event.
   */
  markerselect: AnnotatorEventHandler[] = [];
  /**
   * Event handlers for the `markerdeselect` event.
   */
  markerdeselect: AnnotatorEventHandler[] = [];
  /**
   * Event handlers for the `markercreating` event.
   */
  markercreating: AnnotatorEventHandler[] = [];
  /**
   * Event handlers for the `markercreate` event.
   */
  markercreate: AnnotatorEventHandler[] = [];
  /**
   * Event handlers for the `markerbeforedelete` event.
   */
  markerbeforedelete: AnnotatorEventHandler[] = [];
  /**
   * Event handlers for the `markerdelete` event.
   */
  markerdelete: AnnotatorEventHandler[] = [];
  /**
   * Event handlers for the `markerchange` event.
   * 
   * @since 2.23.0
   */
  markerchange: AnnotatorEventHandler[] = [];
  /**
   * Event handlers for the `focus` event.
   * 
   * @since 2.19.0
   */
  focus: AnnotatorAreaEventHandler[] = [];
  /**
   * Event handlers for the `blur` event.
   * 
   * @since 2.19.0
   */
  blur: AnnotatorAreaEventHandler[] = [];


  /**
   * Add an event handler for a specific event type.
   * @param eventType - event type.
   * @param handler - function to handle the event.
   */
  public addEventListener<T extends keyof IEventListenerRepository>(
    eventType: T,
    handler: EventHandler<T>
  ): void {
    (<Array<EventHandler<T>>>this[eventType]).push(handler);
  }

  /**
   * Remove an event handler for a specific event type.
   * @param eventType - event type.
   * @param handler - function currently handling the event.
   */
  public removeEventListener<T extends keyof IEventListenerRepository>(
    eventType: T,
    handler: EventHandler<T>
  ): void {
    const index = (<Array<EventHandler<T>>>this[eventType]).indexOf(handler);
    if (index > -1) {
      (<Array<EventHandler<T>>>this[eventType]).splice(index, 1);
    }
  }
}
