// @ts-ignore
import emailjs from "@emailjs/browser";
import {
    getStorage, ref as firebaseRef, uploadString, getDownloadURL,
    // @ts-ignore
} from "firebase/storage";
// @ts-ignore
import { getDatabase, ref, child, get, set } from "firebase/database";

import { AnnotatorArea } from "../AnnotationArea";

const constraints: MediaStreamConstraints = {
    video: true,
    preferCurrentTab: true,
};

export class Annotate {
    private async takeScreenShot(img: HTMLImageElement): Promise<string> {
        try {
            const stream: MediaStream = await navigator.mediaDevices.getDisplayMedia(constraints);

            const track: MediaStreamTrack = stream.getVideoTracks()[0];

            const imageCapture: any = new (window as any).ImageCapture(track);

            const bitmap: ImageBitmap = await imageCapture.grabFrame();

            track.stop();

            const canvas: HTMLCanvasElement = document.createElement("canvas");

            canvas.width = bitmap.width;
            canvas.height = bitmap.height;

            const context: CanvasRenderingContext2D | null = canvas.getContext("2d");

            if (context) {
                context.drawImage(bitmap, 0, 0, bitmap.width, bitmap.height);

                const image: string = canvas.toDataURL();

                img.src = image;

                return image;
            } else {
                throw new Error("Unable to get 2D context from canvas");
            }
        } catch (error) {
            console.error("Error capturing screenshot:", error);
            throw error;
        }
    }

    private showAnnotatorArea(target: HTMLImageElement, targetRoot: HTMLImageElement | HTMLElement): void {
        const markerArea = new AnnotatorArea(target);
        markerArea.uiStyleSettings.redoButtonVisible = true
        markerArea.targetRoot = targetRoot;
        markerArea.settings.displayMode = "popup";
        markerArea.uiStyleSettings.zIndex = "1000";
        markerArea.uiStyleSettings.canvasBackgroundColor = "rgba(0,0,0,0.5)";
        markerArea.addEventListener("render", (event) => {
            target.src = event.dataUrl;
        });
        markerArea.show();
    }

    private downloadImage(byteString: string, fileName: string): void {
        const link = document.createElement("a");
        link.download = fileName;
        link.href = byteString;
        document.body.appendChild(link);
        link.click();
        document.body.removeChild(link);
        URL.revokeObjectURL(byteString);
    }

    private async getRecipients(database: any): Promise<HTMLDivElement> {
        const dbRef = ref(database);

        const snapshot = await get(child(dbRef, `emails`));

        const recipientsDiv = document.createElement("div");

        recipientsDiv.id = "recipientsDiv";
        recipientsDiv.style.display = "flex";
        recipientsDiv.style.flexDirection = "column";
        recipientsDiv.style.justifyContent = "center";
        recipientsDiv.style.marginTop = "10px";
        recipientsDiv.style.marginBottom = "10px";
        recipientsDiv.style.padding = "10px";
        recipientsDiv.style.border = "1px solid #ccc";
        recipientsDiv.style.borderRadius = "5px";
        recipientsDiv.style.rowGap = "10px";
        recipientsDiv.style.maxHeight = "200px";
        recipientsDiv.style.overflowY = "scroll";
        const recipientsList = document.createElement("ul");
        recipientsList.id = "recipientList";

        if (snapshot.exists()) {
            const recipients = snapshot.val();

            recipients.forEach((recipient) => {
                const listItem = document.createElement("div");
                listItem.style.textAlign = "left";

                const checkbox = document.createElement("input");
                checkbox.type = "checkbox";
                checkbox.id = "recipient";
                checkbox.style.cursor = "pointer";
                checkbox.onclick = (e) => {
                };
                const label = document.createElement("label");
                label.htmlFor = recipient;
                label.innerHTML = recipient;

                listItem.appendChild(checkbox);
                listItem.appendChild(label);
                recipientsList.appendChild(listItem);
            });
        } else {
            console.log("No data available");
        }
        recipientsDiv.appendChild(recipientsList);
        return recipientsDiv;
    }

    private async sendEmail(storage, templateId, serviceId, emailJsApiKey): Promise<void> {
        const sub: string | null | undefined = (document.getElementById("subject") as HTMLInputElement).value;
        const msg: string | null | undefined = (document.getElementById("message") as HTMLInputElement).value;
        const img: HTMLElement | null = document.getElementById("annotate_preview");

        let recipients: HTMLCollection | undefined = document.getElementById("recipientList")?.children;


        // Check if recipients is not null or undefined before using it
        if (recipients) {
            // Now you can use HTMLCollection methods
            const firstRecipient: Element | null = recipients.item(0);
            const namedRecipient: Element | null = recipients.namedItem("someName");
        }

        // If you still need to convert it to an array
        const recipientsArray: Element[] = Array.from(recipients || []);

        const imageId = Date.now().toString();

        const storageRef = firebaseRef(storage, "images/" + imageId + ".jpg");

        (document.getElementById("sendEmail") as HTMLElement).innerHTML = "Sending";

        const snapshot = await uploadString(storageRef, (img as HTMLImageElement).src, "data_url");
        const url = await getDownloadURL(snapshot.ref);

        const promises = recipientsArray.map((recipient) => {
            if ((recipient.children.item(0) as HTMLInputElement).checked) {
                const email = recipient.children.item(1)?.innerHTML;
                return emailjs.send(
                    serviceId,
                    templateId,
                    {
                        to_email: email,
                        subject: sub,
                        message: msg,
                        src: url,
                    },
                    emailJsApiKey
                );
            }
        });

        Promise.all(promises)
            .then((res) => {
                console.log(res);
                (document.getElementById("sendEmail") as HTMLElement).innerHTML = "Send Mail";
                alert("Email has been sent");
                (document.getElementById("emailDiv") as HTMLElement).style.display = "none";
                (document.getElementById("annotator") as HTMLElement).style.display = "none";
            })
            .catch((e) => {
                alert("There was an error sending the mail");
                console.log(e);
            });
    }

    public async annotate(app_instance: any, templateId: string, serviceId: string, emailJsApiKey: string) {
        let storage: any;
        let database: any;
        if (app_instance) {
            storage = getStorage(app_instance);
            database = getDatabase(app_instance);
        }

        const root = document.querySelector("html");

        const preMain = document.getElementById("main");
        preMain?.remove();

        const main = document.createElement("div");
        main.id = "main";

        var tool = document.createElement("div");

        tool.style.position = "fixed";
        tool.style.top = "0px";
        tool.style.left = "0px";
        tool.style.width = "50px";
        tool.style.height = "50px";
        tool.style.zIndex = "1000";
        tool.style.borderRadius = "50px";
        tool.style.backgroundColor = "green";
        tool.style.margin = "10px";
        tool.style.zIndex = "1000";

        const annotatorModal = document.createElement("div");

        annotatorModal.style.display = "none";
        annotatorModal.style.position = "fixed";
        annotatorModal.style.top = "0";
        annotatorModal.style.left = "0";
        annotatorModal.style.width = "100%";
        annotatorModal.style.height = "100%";
        annotatorModal.style.justifyContent = "center";
        annotatorModal.style.alignItems = "center";
        annotatorModal.style.backgroundColor = "rgba(0,0,0,0.5)";
        annotatorModal.style.overflow = "auto";
        annotatorModal.style.zIndex = "10000";
        annotatorModal.id = "annotator";

        const annotator = document.createElement("div");
        annotator.style.display = "flex";
        annotator.style.flexDirection = "column";
        annotator.style.justifyContent = "center";
        annotator.style.width = "70%";
        annotator.style.height = "fit-content";
        annotator.style.marginLeft = "auto";
        annotator.style.marginRight = "auto";
        annotator.style.backgroundColor = "white";
        annotator.style.padding = "10px";
        annotator.style.borderRadius = "10px";
        annotator.style.boxShadow = "0 0 10px rgba(0,0,0,0.5)";
        annotator.style.textAlign = "center";
        annotator.style.overflowY = "scroll";
        annotator.style.zIndex = "1000";

        const emailDiv = document.createElement("div");
        emailDiv.id = "emailDiv";
        emailDiv.style.marginTop = "20px";
        emailDiv.style.display = "none";
        emailDiv.style.flexDirection = "column";
        emailDiv.style.justifyContent = "center";
        emailDiv.style.width = "100%";

        const recipient: HTMLInputElement = document.createElement("input");
        recipient.type = "email";
        recipient.id = "recipient";
        recipient.placeholder = "Add a new recipient";
        recipient.style.marginBottom = "10px";
        recipient.style.padding = "10px";
        recipient.style.borderRadius = "5px";
        recipient.style.border = "1px solid #ccc";

        const addRecipientButton = document.createElement("button");
        addRecipientButton.id = "addRecipient";
        addRecipientButton.innerHTML = "Add New Recipient";
        addRecipientButton.style.width = "30%";
        addRecipientButton.style.cursor = "pointer";
        addRecipientButton.onclick = async () => {
            const recipient2 = (document.getElementById("recipient") as HTMLInputElement).value;
            if (!recipient2) {
                alert("Please enter a valid email address");
                return;
            }
            let listElement: HTMLElement | undefined | null = document.getElementById("recipientList");
            let list: HTMLCollection | undefined = listElement?.children;
            // Check if recipients is not null or undefined before using it
            if (list) {
                // Now you can use HTMLCollection methods
                const firstRecipient: Element | null = list.item(0);
                const namedRecipient: Element | null = list.namedItem("someName");
            }

            // If you still need to convert it to an array
            const recipientsArray: Element[] = Array.from(list || []);

            let listArr: (string | undefined)[] = recipientsArray.map((item) => item.children.item(1)?.innerHTML);
            if (listArr.includes(recipient2)) {
                alert("Recipient already exists");
                return;
            }
            let dbRef: any;
            if (database) {
                dbRef = ref(database);
            }


            addRecipientButton.innerHTML = "Adding";
            if (dbRef) {
                await set(child(dbRef, `emails`), [...listArr, recipient]);
            }

            const listItem = document.createElement("div");
            listItem.style.textAlign = "left";

            const checkbox = document.createElement("input");
            checkbox.type = "checkbox";
            checkbox.id = "recipient";
            checkbox.style.cursor = "pointer";
            checkbox.onclick = (e) => {
            };
            const label = document.createElement("label");
            label.htmlFor = recipient.id;
            label.innerHTML = recipient.placeholder;

            listItem.appendChild(checkbox);
            listItem.appendChild(label);
            listElement?.appendChild(listItem);

            (document.getElementById("recipient") as HTMLInputElement).value = "";
            addRecipientButton.innerHTML = "Add New Recipient";
        };

        const subject = document.createElement("input");
        subject.id = "subject";
        subject.type = "text";
        subject.placeholder = "Subject";
        subject.style.marginBottom = "10px";
        subject.style.padding = "10px";
        subject.style.borderRadius = "5px";
        subject.style.border = "1px solid #ccc";

        const message = document.createElement("textarea");
        message.id = "message";
        message.placeholder = "Message";
        message.style.marginBottom = "10px";
        message.style.padding = "10px";
        message.style.borderRadius = "5px";
        message.style.border = "1px solid #ccc";
        message.style.height = "200px";

        const sendEmailButton = document.createElement("button");
        sendEmailButton.id = "sendEmail";
        sendEmailButton.innerHTML = "Send Email";
        sendEmailButton.style.width = "30%";
        sendEmailButton.style.cursor = "pointer";
        sendEmailButton.onclick = async () => {
            if (storage && templateId && serviceId && emailJsApiKey) {
                this.sendEmail(storage, templateId, serviceId, emailJsApiKey);
            }
        };

        //   const recipeintsDiv = await getRecipients(database);
        emailDiv.appendChild(recipient);
        emailDiv.appendChild(addRecipientButton);
        //   emailDiv.appendChild(recipeintsDiv);

        emailDiv.appendChild(subject);
        emailDiv.appendChild(message);

        const emailButtonDiv = document.createElement("div");
        emailButtonDiv.style.display = "flex";
        emailButtonDiv.style.flexDirection = "row";
        emailButtonDiv.style.gap = "10px";
        emailButtonDiv.style.marginTop = "10px";

        const closeEmailDivButton = document.createElement("button");
        closeEmailDivButton.id = "closeEmailDivButton";
        closeEmailDivButton.innerHTML = "Close";
        closeEmailDivButton.style.marginRight = "auto";
        closeEmailDivButton.style.width = "30%";
        closeEmailDivButton.style.cursor = "pointer";
        closeEmailDivButton.onclick = () => {
            emailDiv.style.display = "none";
        };

        emailButtonDiv.appendChild(sendEmailButton);
        emailButtonDiv.appendChild(closeEmailDivButton);

        emailDiv.appendChild(emailButtonDiv);

        const buttonsDiv = document.createElement("div");
        buttonsDiv.style.display = "flex";
        buttonsDiv.style.flexDirection = "row";
        buttonsDiv.style.justifyContent = "space-between";
        buttonsDiv.style.width = "50%";
        buttonsDiv.style.marginTop = "10px";
        buttonsDiv.style.padding = "10px";
        buttonsDiv.style.columnGap = "10px";
        buttonsDiv.style.marginLeft = "auto";

        root?.appendChild(main);

        const img = document.createElement("img");
        img.id = "annotate_preview";
        img.style.cursor = "pointer";

        img.onclick = (e) => {
            this.showAnnotatorArea(img, annotator);
        };

        const openButton = document.createElement("button");
        openButton.id = "open";
        openButton.innerHTML = '<svg xmlns="http://www.w3.org/2000/svg" width="16" height="16" fill="currentColor" class="bi bi-camera" viewBox="0 0 16 16"><path d="M15.5 2h-2.2l-1.3-1.3c-.1-.1-.2-.1-.3 0l-1.3 1.3H5.8l-1.3-1.3c-.1-.1-.2-.1-.3 0L2.7 2H.5c-.3 0-.5.2-.5.5v12c0 .3.2.5.5.5h15c.3 0 .5-.2.5-.5v-12c0-.3-.2-.5-.5-.5zM8 11.8c-2.1 0-3.8-1.7-3.8-3.8s1.7-3.8 3.8-3.8 3.8 1.7 3.8 3.8-1.7 3.8-3.8 3.8zm0-6.5c-1.5 0-2.7 1.2-2.7 2.7s1.2 2.7 2.7 2.7 2.7-1.2 2.7-2.7-1.2-2.7-2.7-2.7z"/></svg>';
        openButton.style.width = "50px";
        openButton.style.height = "50px";
        openButton.style.height = "50px";
        openButton.style.borderRadius = "50px";
        // Center the SVG vertically using Flexbox
        openButton.style.display = "flex";
        openButton.style.alignItems = "center";
        openButton.style.justifyContent = "center";
        openButton.style.border = "none";
        openButton.style.color = "white";
        openButton.style.cursor = "pointer";
        openButton.style.backgroundImage = "linear-gradient(to right, purple, blue)";


        tool.appendChild(openButton);
        document.body.appendChild(tool);

        const closeButton = document.createElement("button");
        closeButton.id = "close";
        closeButton.innerHTML = "Close";
        closeButton.style.marginLeft = "auto";
        closeButton.style.width = "50%";
        closeButton.style.cursor = "pointer";

        const downloadButton = document.createElement("button");
        downloadButton.id = "download";
        downloadButton.innerHTML = "Download";
        downloadButton.style.marginRight = "auto";
        downloadButton.style.width = "50%";
        downloadButton.style.cursor = "pointer";
        downloadButton.onclick = () => {
            this.downloadImage(img.src, "screenshot.png");
        };

        const emailButton = document.createElement("button");
        emailButton.id = "emailButton";
        emailButton.innerHTML = "Email";
        emailButton.style.marginRight = "left";
        emailButton.style.cursor = "pointer";
        emailButton.style.width = "50%";
        emailButton.onclick = () => {

            if (app_instance && emailJsApiKey && templateId && serviceId) {

                emailDiv.style.display = "flex";
            } else {
                alert("Please provide appropriate data for email feature.")
            }
        };

        const openImage = () => {
            this.takeScreenShot(img)
                .then(() => {
                    annotatorModal.style.display = "flex";
                })
                .catch((e) => {
                    annotatorModal.style.display = "none";
                    tool.style.visibility = "visible";
                });
        };

        openButton.onclick = () => {
            openImage();
            tool.style.visibility = "hidden";
        };

        closeButton.onclick = () => {
            annotatorModal.style.display = "none";
            tool.style.visibility = "visible";
        };


        buttonsDiv.appendChild(emailButton);
        buttonsDiv.appendChild(downloadButton);
        buttonsDiv.appendChild(closeButton);

        annotator.appendChild(img);
        annotator.appendChild(emailDiv);
        annotator.appendChild(buttonsDiv);

        annotatorModal.appendChild(annotator);

        main.appendChild(annotatorModal);
        let pressedKeys: string[] = [];

        document.addEventListener("keydown", handleKeydown);
        document.addEventListener("keyup", handleKeyup);

        function handleKeydown(e: KeyboardEvent) {
            if (!pressedKeys.includes(e.key)) {
                pressedKeys.push(e.key);

                if (
                    pressedKeys.includes("Control") &&
                    pressedKeys.includes("b") &&
                    pressedKeys.length == 2
                ) {
                    e.preventDefault();
                    openImage();
                    tool.style.visibility = "hidden";

                    return;
                }
            }
        }

        function handleKeyup(e) {
            const index = pressedKeys.indexOf(e.key);
            pressedKeys.splice(index, 1);
        }

        return "";
    }

}

